#
# Copyright 2022 The AI Flow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import time
import socket
from collections import Iterable
from typing import Union, Tuple

from notification_service.model.event import Event, ANY_CONDITION
from notification_service.model.sender_event_count import SenderEventCount
from notification_service.storage.event_storage import BaseEventStorage

from mongoengine import connect
from mongoengine import Document, StringField, SequenceField, LongField, BooleanField


class MongoEventStorage(BaseEventStorage):
    def __init__(self, *args, **kwargs):
        self.db_conn = self.setup_connection(**kwargs)
        self.db = None
        self.server_ip = socket.gethostbyname(socket.gethostname())

    def setup_connection(self, **kwargs):
        db_conf = {
            "host": kwargs.get("host"),
            "port": kwargs.get("port"),
            "db": kwargs.get("db"),
        }
        self.db = db_conf.get('db')
        username = kwargs.get("username", None)
        password = kwargs.get("password", None)
        authentication_source = kwargs.get("authentication_source", "admin")
        if (username or password) and not (username and password):
            raise Exception("Please provide valid username and password")
        if username and password:
            db_conf.update({
                "username": username,
                "password": password,
                "authentication_source": authentication_source
            })
        return connect(**db_conf)

    def add_event(self, event: Event, uuid: str):
        kwargs = {
            "server_ip": self.server_ip,
            "create_time": int(time.time() * 1000),
            "key": event.key,
            "value": event.value,
            "context": event.context,
            "namespace": event.namespace,
            "sender": event.sender,
            "uuid": uuid
        }
        mongo_event = MongoEvent(**kwargs)
        mongo_event.save()
        mongo_event.reload()
        event.create_time = mongo_event.create_time
        event.offset = mongo_event.offset
        return event

    def list_events(self,
                    key: str = None,
                    namespace: str = None,
                    sender: str = None,
                    begin_offset: int = None,
                    end_offset: int = None):
        key = None if key == "" else key
        namespace = None if namespace == "" else namespace
        sender = None if sender == "" else sender
        res = MongoEvent.get_base_events(key, namespace, sender, begin_offset, end_offset)
        return res

    def count_events(self,
                     key: str = None,
                     namespace: str = None,
                     sender: str = None,
                     begin_offset: int = None,
                     end_offset: int = None):
        key = None if key == "" else key
        namespace = None if namespace == "" else namespace
        sender = None if sender == "" else sender
        res = MongoEvent.count_base_events(key, namespace, sender, begin_offset, end_offset)
        return res

    def list_all_events(self, start_time: int):
        res = MongoEvent.get_base_events_by_time(start_time)
        return res

    def list_all_events_from_offset(self, start_offset: int, end_offset: int = None):
        res = MongoEvent.get_base_events_by_offset(start_offset, end_offset)
        return res

    def register_client(self, namespace: str = None, sender: str = None) -> int:
        client = MongoClientModel(namespace=namespace, sender=sender, create_time=int(time.time() * 1000))
        client.save()
        return client.id

    def delete_client(self, client_id):
        client_to_delete = MongoClientModel.objects(id=client_id).first()
        if client_to_delete is None:
            raise Exception("You are trying to delete an non-existing notification client!")
        client_to_delete.is_deleted = True
        client_to_delete.save()

    def is_client_exists(self, client_id) -> bool:
        client_to_check = MongoClientModel.objects(id=client_id, is_deleted__ne=True).first()
        return client_to_check is not None

    def get_event_by_uuid(self, uuid: str):
        return MongoEvent.get_event_by_uuid(uuid)

    def timestamp_to_event_offset(self, timestamp: int) -> int:
        return MongoEvent.timestamp_to_event_offset(timestamp=timestamp)

    def clean_up(self):
        MongoEvent.delete_by_client(self.server_ip)
        if self.db is not None:
            self.db_conn.drop_database(self.db)


class MongoEvent(Document):
    server_ip = StringField()   # track server that the event belongs to
    key = StringField()
    value = StringField()
    offset = SequenceField()  # use 'offset' as the auto increase id
    create_time = LongField()
    context = StringField()
    namespace = StringField()
    sender = StringField()
    uuid = StringField()

    def to_dict(self):
        return {
            "key": self.key,
            "value": self.value,
            "offset": int(self.offset),
            "create_time": self.create_time,
            "context": self.context,
            "namespace": self.namespace,
            "sender": self.sender
        }

    def to_base_event(self):
        base_event = Event(**self.to_dict())
        return base_event

    @classmethod
    def convert_to_base_events(cls, mongo_events: list = None):
        if not mongo_events:
            return []
        base_events = []
        for mongo_event in mongo_events:
            base_events.append(mongo_event.to_base_event())
        return base_events

    @classmethod
    def convert_to_event_counts(cls, mongo_counts: list = None):
        if not mongo_counts:
            return []
        event_counts = []
        for mongo_count in mongo_counts:
            event_counts.append(SenderEventCount(sender=mongo_count.sender, event_count=mongo_count.count))
        return event_counts

    @classmethod
    def get_base_events_by_offset(cls, start_offset: int, end_offset: int = None):
        conditions = dict()
        conditions["offset__gt"] = start_offset
        if end_offset is not None and end_offset > 0:
            conditions["offset__lte"] = end_offset
        mongo_events = cls.objects(**conditions).order_by("offset")
        return cls.convert_to_base_events(mongo_events)

    @classmethod
    def get_base_events(cls,
                        key: str = None,
                        namespace: str = None,
                        sender: str = None,
                        begin_offset: int = None,
                        end_offset: int = None):
        conditions = dict()
        if key and key != ANY_CONDITION:
            conditions["key"] = key
        if namespace and namespace != ANY_CONDITION:
            conditions["namespace"] = namespace
        if sender and sender != ANY_CONDITION:
            conditions["sender"] = sender
        if begin_offset:
            conditions["offset__gt"] = begin_offset
        if end_offset:
            conditions["offset__lte"] = end_offset

        mongo_events = cls.objects(**conditions).order_by("offset")
        return cls.convert_to_base_events(mongo_events)

    @classmethod
    def count_base_events(cls,
                          key: str = None,
                          namespace: str = None,
                          sender: str = None,
                          begin_offset: int = None,
                          end_offset: int = None):
        conditions = dict()
        if key and key != ANY_CONDITION:
            conditions["key"] = key
        if namespace and namespace != ANY_CONDITION:
            conditions["namespace"] = namespace
        if sender and sender != ANY_CONDITION:
            conditions["sender"] = sender
        if begin_offset:
            conditions["offset__gt"] = begin_offset
        if end_offset:
            conditions["offset__lte"] = end_offset
        mongo_counts = cls.objects(**conditions).aggregate([{"$group": {"_id": "$sender", "count": {"$sum": 1}}}])
        return cls.convert_to_event_counts(mongo_counts)

    @classmethod
    def get_base_events_by_time(cls, create_time: int = None):
        mongo_events = cls.objects(create_time__gte=create_time).order_by("offset")
        return cls.convert_to_base_events(mongo_events)

    @classmethod
    def get_by_key(cls, key: str = None, start: int = None, end: int = None, sort_key: str = "offset"):
        if key:
            return cls.objects(key=key).order_by(sort_key)[start:end]
        else:
            raise Exception("key is empty, please provide valid key")

    @classmethod
    def delete_by_client(cls, server_ip):
        cls.objects(server_ip=server_ip).delete()

    @classmethod
    def get_event_by_uuid(cls, uuid):
        event = cls.objects(uuid=uuid).first()
        return event.to_base_event()

    @classmethod
    def timestamp_to_event_offset(cls, timestamp):
        event = cls.objects(create_time__lte=timestamp).order_by('-offset').first()
        return event.offset if event else 0


class MongoClientModel(Document):
    id = SequenceField(primary_key=True)
    namespace = StringField()
    sender = StringField()
    create_time = LongField()
    is_deleted = BooleanField(default=False)

    def __repr__(self):
        return '<Document Client ({}, {}, {}, {})>'.format(
            self.id,
            self.namespace,
            self.sender,
            self.is_deleted)



# Copyright 2022 The AI Flow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""add_tables

Revision ID: 87cb292bcc31
Revises: 
Create Date: 2021-10-15 12:04:11.696836

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '87cb292bcc31'
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    op.create_table('event_model',
                    sa.Column('offset', sa.BigInteger().with_variant(sa.Integer(), 'sqlite'), nullable=False),
                    sa.Column('key', sa.String(length=1024), nullable=False),
                    sa.Column('value', sa.Text(), nullable=True),
                    sa.Column('context', sa.Text(), nullable=True),
                    sa.Column('namespace', sa.String(length=1024), nullable=True),
                    sa.Column('sender', sa.String(length=1024), nullable=True),
                    sa.Column('create_time', sa.BigInteger(), nullable=False),
                    sa.Column('uuid', sa.String(length=40), nullable=False),
                    sa.PrimaryKeyConstraint('offset'),
                    sa.UniqueConstraint('uuid')
                    )
    op.create_table('notification_client',
                    sa.Column('id', sa.BigInteger().with_variant(sa.Integer(), 'sqlite'), autoincrement=True,
                              nullable=False),
                    sa.Column('namespace', sa.String(length=1024), nullable=True),
                    sa.Column('sender', sa.String(length=1024), nullable=True),
                    sa.Column('create_time', sa.BigInteger(), nullable=True),
                    sa.Column('is_deleted', sa.Boolean(), nullable=True),
                    sa.PrimaryKeyConstraint('id')
                    )


def downgrade():
    op.drop_table('notification_client')
    op.drop_table('event_model')



#
# Copyright 2022 The AI Flow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

import time

from notification_service.model.member import Member
from sqlalchemy import Column, String, BigInteger, Integer

from notification_service.util import db
from notification_service.util.db import provide_session

from notification_service.storage.alchemy.base import Base
from notification_service.storage.high_availability import HighAvailabilityStorage


class DbHighAvailabilityStorage(HighAvailabilityStorage):

    def __init__(self, db_conn=None):
        if db_conn is not None:
            db.SQL_ALCHEMY_CONN = db_conn
        db.prepare_db()

    def list_living_members(self, ttl_ms):
        return MemberModel.get_living_members(ttl_ms)

    def list_dead_members(self, ttl_ms):
        return MemberModel.get_dead_members(ttl_ms)

    def update_member(self, server_uri, server_uuid):
        MemberModel.update_member(server_uri, server_uuid)

    def delete_member(self, server_uri=None, server_uuid=None):
        MemberModel.delete_member(server_uri, server_uuid)

    def clear_dead_members(self, ttl_ms):
        MemberModel.clear_dead_members(ttl_ms)


class MemberModel(Base):
    __tablename__ = "member_model"
    id = Column(BigInteger().with_variant(Integer, "sqlite"), primary_key=True)
    version = Column(BigInteger(), nullable=False)
    server_uri = Column(String(767), nullable=False, unique=True)
    update_time = Column(BigInteger(), nullable=False)
    uuid = Column(String(128), nullable=False, unique=True)

    @staticmethod
    @provide_session
    def update_member(server_uri, server_uuid, session=None):
        member = session.query(MemberModel) \
            .filter(MemberModel.server_uri == server_uri).first()
        if member is None:
            member = MemberModel()
            member.version = 1
            member.server_uri = server_uri
            member.update_time = time.time_ns() / 1000000
            member.uuid = server_uuid
            session.add(member)
        else:
            if member.uuid != server_uuid:
                raise Exception("The server uri '%s' is already exists in the storage!" %
                                server_uri)
            member.version += 1
            member.update_time = time.time_ns() / 1000000
        session.commit()

    @staticmethod
    @provide_session
    def get_living_members(ttl, session=None):
        member_models = session.query(MemberModel) \
            .filter(MemberModel.update_time >= time.time_ns() / 1000000 - ttl) \
            .all()
        return [Member(m.version, m.server_uri, int(m.update_time)) for m in member_models]

    @staticmethod
    @provide_session
    def get_dead_members(ttl, session=None):
        member_models = session.query(MemberModel) \
            .filter(MemberModel.update_time < time.time_ns() / 1000000 - ttl) \
            .all()
        return [Member(m.version, m.server_uri, int(m.update_time)) for m in member_models]

    @staticmethod
    @provide_session
    def delete_member(server_uri=None, server_uuid=None, session=None):
        conditions = []
        if server_uri:
            conditions.append(MemberModel.server_uri == server_uri)
        if server_uuid:
            conditions.append(MemberModel.uuid == server_uuid)
        if len(conditions) != 1:
            raise Exception("Please provide exactly one param, server_uri or server_uuid")
        member = session.query(MemberModel).filter(*conditions).first()
        if member is not None:
            session.delete(member)
        session.commit()

    @staticmethod
    @provide_session
    def clear_dead_members(ttl, session=None):
        session.query(MemberModel) \
            .filter(MemberModel.update_time < time.time_ns() / 1000000 - ttl) \
            .delete()
        session.commit()



#
# Copyright 2022 The AI Flow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import logging
import os

import grpc
from notification_service.server.ha_manager import SimpleNotificationServerHaManager
from notification_service.server.server import NotificationServer

from notification_service.storage.alchemy.db_event_storage import DbEventStorage
from notification_service.storage.alchemy.db_high_availability_storage import DbHighAvailabilityStorage
from notification_service.util.server_config import NotificationServerConfig
from notification_service.rpc.service import NotificationService, HighAvailableNotificationService
from notification_service.util.utils import get_ip_addr


logger = logging.getLogger(__name__)


class NotificationServerRunner(object):
    def __init__(self, config_file):
        if not os.path.exists(config_file):
            raise IOError('Config file {} not exist!'.format(config_file))
        logger.info("Loading Notification server config from path: {}".format(config_file))
        self.config = NotificationServerConfig(config_file=config_file)

    def _init_server(self):
        if self.config.db_uri:
            self.storage = DbEventStorage(self.config.db_uri)
        else:
            raise Exception('Failed to start notification service without database connection info.')
        if self.config.enable_ha:
            server_uri = self.config.advertised_uri \
                if self.config.advertised_uri is not None else get_ip_addr() + ':' + str(self.config.port)
            ha_storage = DbHighAvailabilityStorage(db_conn=self.config.db_uri)
            ha_manager = SimpleNotificationServerHaManager()
            service = HighAvailableNotificationService(
                self.storage,
                ha_manager,
                server_uri,
                ha_storage,
                5000)
            self.server = NotificationServer(service=service,
                                             port=int(self.config.port))
        else:
            self.server = NotificationServer(service=NotificationService(self.storage),
                                             port=int(self.config.port))

    def start(self, is_block=False):
        logger.info("Starting notification server with config: {}".format(self.config))
        self._init_server()
        self.server.run(is_block)
        if not is_block:
            self._wait_for_server_available(timeout=self.config.wait_for_server_started_timeout)

    def stop(self):
        self.server.stop()

    def _wait_for_server_available(self, timeout):
        """
        Wait for server to be started and available.

        :param timeout: Float value. Seconds to wait for server available.
                        If None, wait forever until server started.
        """
        server_uri = 'localhost:{}'.format(self.config.port)
        try:
            channel = grpc.insecure_channel(server_uri)
            grpc.channel_ready_future(channel).result(timeout=timeout)
            logging.info("Notification Server started successfully.")
        except grpc.FutureTimeoutError as e:
            logging.error('Notification Server is not available after waiting for {} seconds.'.format(timeout))
            raise e



#
# Copyright 2022 The AI Flow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
"""Utilities module for cli"""
import os
import sys

from pygments.formatters.terminal import TerminalFormatter
from pygments.formatters.terminal256 import Terminal256Formatter


def get_terminal_formatter(**opts):
    """Returns the best formatter available in the current terminal."""
    if '256' in os.environ.get('TERM', ''):
        formatter = Terminal256Formatter(**opts)
    else:
        formatter = TerminalFormatter(**opts)
    return formatter


class ColorMode:
    """Coloring modes. If `auto` is then automatically detected."""

    ON = "on"
    OFF = "off"
    AUTO = "auto"


def is_tty():
    """
    Checks if the standard output is connected (is associated with a terminal device) to a tty(-like)
    device.
    """
    if not hasattr(sys.stdout, "isatty"):
        return False
    return sys.stdout.isatty()


def is_terminal_support_colors() -> bool:
    """Try to determine if the current terminal supports colors."""
    if sys.platform == "win32":
        return False
    if not is_tty():
        return False
    if "COLORTERM" in os.environ:
        return True
    term = os.environ.get("TERM", "dumb").lower()
    if term in ("xterm", "linux") or "color" in term:
        return True
    return False


def should_use_colors(args) -> bool:
    """Processes arguments and decides whether to enable color in output"""
    if args.color == ColorMode.ON:
        return True
    if args.color == ColorMode.OFF:
        return False
    return is_terminal_support_colors()



# coding:utf-8
#
# Copyright 2022 The AI Flow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import os
import sys
import time

from setuptools import setup, find_packages

CURRENT_DIR = os.path.abspath(os.path.dirname(__file__))
nightly_build = os.getenv('NIGHTLY_BUILD') == 'true'


def get_script():
    bin_dir = os.path.join(CURRENT_DIR, "bin")
    return [os.path.join("bin", filename) for filename in os.listdir(bin_dir)]


version_file = os.path.join(CURRENT_DIR, 'notification_service/version.py')
try:
    exec(open(version_file).read())
except IOError:
    print("Failed to load notification_service version file for packaging. " +
          "'%s' not found!" % version_file,
          file=sys.stderr)
    sys.exit(-1)
VERSION = time.strftime('%Y.%m.%d', time.localtime(time.time())) if nightly_build else __version__ # noqa
PACKAGE_NAME = 'notification_service_nightly' if nightly_build else 'notification_service'


require_file = '{}/{}'.format(os.path.dirname(os.path.abspath(__file__)), "requirements.txt")
with open(require_file) as f:
    context = f.read()
    require_file_lines = context.strip().split('\n')

require_packages = []
for line in require_file_lines:
    if not len(line.strip()) == 0 and not line.startswith("#"):
        require_packages.append(line)

setup(
    name=PACKAGE_NAME,
    version=VERSION,
    description='A Python package which provides stable notification service.',
    author='',
    author_email='flink.aiflow@gmail.com',
    url='https://github.com/flink-extended/ai-flow',
    packages=find_packages(exclude=['tests*']),
    scripts=get_script(),
    package_data={'notification_service': ['alembic.ini']},
    include_package_data=True,
    install_requires=require_packages,
    entry_points={
        'console_scripts': [
            'notification = notification_service.__main__:main'
        ]
    }
)



# Copyright 2022 The AI Flow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
import time
from sqlalchemy import Column, String, BigInteger, Integer, Boolean

from notification_service.util.db import provide_session
from notification_service.storage.alchemy.base import Base


class ClientModel(Base):
    __tablename__ = "notification_client"
    id = Column(BigInteger().with_variant(Integer, "sqlite"), autoincrement=True, primary_key=True)
    namespace = Column(String(1024))
    sender = Column(String(1024))
    create_time = Column(BigInteger)
    is_deleted = Column(Boolean, default=False)

    @staticmethod
    @provide_session
    def register_client(namespace: str = None, sender: str = None, session=None):
        client = ClientModel()
        client.namespace = namespace
        client.sender = sender
        client.create_time = int(time.time() * 1000)
        session.add(client)
        session.commit()
        return client.id

    @staticmethod
    @provide_session
    def delete_client(client_id, session=None):
        client_to_delete = session.query(ClientModel).filter(ClientModel.id == client_id).first()
        if client_to_delete is None:
            raise Exception("You are trying to delete an non-existing notification client!")
        client_to_delete.is_deleted = True
        session.flush()

    @staticmethod
    @provide_session
    def is_client_exists(client_id, session=None):
        client_to_check = session.query(ClientModel).filter(ClientModel.id == client_id,
                                                            ClientModel.is_deleted.is_(False)).first()
        return client_to_check is not None


#
# Copyright 2022 The AI Flow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
from abc import ABC, abstractmethod
from typing import List

from notification_service.model.member import Member


class HighAvailabilityStorage(ABC):

    @abstractmethod
    def list_living_members(self, ttl_ms) -> List[Member]:
        pass

    @abstractmethod
    def update_member(self, server_uri, server_uuid):
        pass

    @abstractmethod
    def clear_dead_members(self, ttl_ms):
        pass


# Copyright 2022 The AI Flow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
import logging.config
import os

from notification_service.util.server_config import NotificationServerConfig

# We hard code the logging config, we should make it configurable in the future.
logging.config.dictConfig({
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'default': {
            'format': '[%(asctime)s - %(filename)s:%(lineno)d [%(threadName)s] - %(levelname)s: %(message)s',
        }
    },
    'handlers': {
        'console': {
            'class': 'logging.StreamHandler',
            'stream': 'ext://sys.stderr',
            'formatter': 'default'
        }
    },
    'root': {
        'level': 'INFO',
        'handlers': ['console']
    }
})


def get_notification_home():
    if 'NOTIFICATION_HOME' in os.environ:
        home = os.getenv('NOTIFICATION_HOME')
    else:
        home = os.getenv('HOME') + '/notification_service'
    return home


NOTIFICATION_HOME = get_notification_home()
NOTIFICATION_PID_FILENAME = 'notification_server.pid'


def get_configuration_file_path(notification_home=None):
    if not notification_home:
        notification_home = NOTIFICATION_HOME
    config_file_path = notification_home + '/notification_server.yaml'
    if not os.path.exists(config_file_path):
        raise FileNotFoundError('Do not find config file {}'.format(config_file_path))
    return config_file_path


def get_configuration():
    config_file_path = get_configuration_file_path()
    return NotificationServerConfig(config_file=config_file_path)



#
# Copyright 2022 The AI Flow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#


# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from . import notification_service_pb2 as notification__service__pb2


class NotificationServiceStub(object):
    """AirFlowService provides notification function rest endpoint of NotificationService for Notification Service component.
    Functions of NotificationService include:
    1.Send event.
    2.List events.
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.sendEvent = channel.unary_unary(
                '/notification_service.NotificationService/sendEvent',
                request_serializer=notification__service__pb2.SendEventRequest.SerializeToString,
                response_deserializer=notification__service__pb2.SendEventsResponse.FromString,
                )
        self.listEvents = channel.unary_unary(
                '/notification_service.NotificationService/listEvents',
                request_serializer=notification__service__pb2.ListEventsRequest.SerializeToString,
                response_deserializer=notification__service__pb2.ListEventsResponse.FromString,
                )
        self.countEvents = channel.unary_unary(
                '/notification_service.NotificationService/countEvents',
                request_serializer=notification__service__pb2.CountEventsRequest.SerializeToString,
                response_deserializer=notification__service__pb2.CountEventsResponse.FromString,
                )
        self.listAllEvents = channel.unary_unary(
                '/notification_service.NotificationService/listAllEvents',
                request_serializer=notification__service__pb2.ListAllEventsRequest.SerializeToString,
                response_deserializer=notification__service__pb2.ListEventsResponse.FromString,
                )
        self.notify = channel.unary_unary(
                '/notification_service.NotificationService/notify',
                request_serializer=notification__service__pb2.NotifyRequest.SerializeToString,
                response_deserializer=notification__service__pb2.CommonResponse.FromString,
                )
        self.listMembers = channel.unary_unary(
                '/notification_service.NotificationService/listMembers',
                request_serializer=notification__service__pb2.ListMembersRequest.SerializeToString,
                response_deserializer=notification__service__pb2.ListMembersResponse.FromString,
                )
        self.notifyNewMember = channel.unary_unary(
                '/notification_service.NotificationService/notifyNewMember',
                request_serializer=notification__service__pb2.NotifyNewMemberRequest.SerializeToString,
                response_deserializer=notification__service__pb2.CommonResponse.FromString,
                )
        self.registerClient = channel.unary_unary(
                '/notification_service.NotificationService/registerClient',
                request_serializer=notification__service__pb2.RegisterClientRequest.SerializeToString,
                response_deserializer=notification__service__pb2.RegisterClientResponse.FromString,
                )
        self.deleteClient = channel.unary_unary(
                '/notification_service.NotificationService/deleteClient',
                request_serializer=notification__service__pb2.ClientIdRequest.SerializeToString,
                response_deserializer=notification__service__pb2.CommonResponse.FromString,
                )
        self.isClientExists = channel.unary_unary(
                '/notification_service.NotificationService/isClientExists',
                request_serializer=notification__service__pb2.ClientIdRequest.SerializeToString,
                response_deserializer=notification__service__pb2.isClientExistsResponse.FromString,
                )
        self.timestampToEventOffset = channel.unary_unary(
                '/notification_service.NotificationService/timestampToEventOffset',
                request_serializer=notification__service__pb2.TimeToOffsetRequest.SerializeToString,
                response_deserializer=notification__service__pb2.TimeToOffsetResponse.FromString,
                )


class NotificationServiceServicer(object):
    """AirFlowService provides notification function rest endpoint of NotificationService for Notification Service component.
    Functions of NotificationService include:
    1.Send event.
    2.List events.
    """

    def sendEvent(self, request, context):
        """Send event.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def listEvents(self, request, context):
        """List events.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def countEvents(self, request, context):
        """Count events.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def listAllEvents(self, request, context):
        """List all events
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def notify(self, request, context):
        """Accepts notifications from other members.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def listMembers(self, request, context):
        """List current living members.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def notifyNewMember(self, request, context):
        """Notify current members that there is a new member added.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def registerClient(self, request, context):
        """Register notification client in the db of notification service
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def deleteClient(self, request, context):
        """Delete notification client
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def isClientExists(self, request, context):
        """Check if a notification client has been registered in the db
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def timestampToEventOffset(self, request, context):
        """Get the event offset by timestamp
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_NotificationServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'sendEvent': grpc.unary_unary_rpc_method_handler(
                    servicer.sendEvent,
                    request_deserializer=notification__service__pb2.SendEventRequest.FromString,
                    response_serializer=notification__service__pb2.SendEventsResponse.SerializeToString,
            ),
            'listEvents': grpc.unary_unary_rpc_method_handler(
                    servicer.listEvents,
                    request_deserializer=notification__service__pb2.ListEventsRequest.FromString,
                    response_serializer=notification__service__pb2.ListEventsResponse.SerializeToString,
            ),
            'countEvents': grpc.unary_unary_rpc_method_handler(
                    servicer.countEvents,
                    request_deserializer=notification__service__pb2.CountEventsRequest.FromString,
                    response_serializer=notification__service__pb2.CountEventsResponse.SerializeToString,
            ),
            'listAllEvents': grpc.unary_unary_rpc_method_handler(
                    servicer.listAllEvents,
                    request_deserializer=notification__service__pb2.ListAllEventsRequest.FromString,
                    response_serializer=notification__service__pb2.ListEventsResponse.SerializeToString,
            ),
            'notify': grpc.unary_unary_rpc_method_handler(
                    servicer.notify,
                    request_deserializer=notification__service__pb2.NotifyRequest.FromString,
                    response_serializer=notification__service__pb2.CommonResponse.SerializeToString,
            ),
            'listMembers': grpc.unary_unary_rpc_method_handler(
                    servicer.listMembers,
                    request_deserializer=notification__service__pb2.ListMembersRequest.FromString,
                    response_serializer=notification__service__pb2.ListMembersResponse.SerializeToString,
            ),
            'notifyNewMember': grpc.unary_unary_rpc_method_handler(
                    servicer.notifyNewMember,
                    request_deserializer=notification__service__pb2.NotifyNewMemberRequest.FromString,
                    response_serializer=notification__service__pb2.CommonResponse.SerializeToString,
            ),
            'registerClient': grpc.unary_unary_rpc_method_handler(
                    servicer.registerClient,
                    request_deserializer=notification__service__pb2.RegisterClientRequest.FromString,
                    response_serializer=notification__service__pb2.RegisterClientResponse.SerializeToString,
            ),
            'deleteClient': grpc.unary_unary_rpc_method_handler(
                    servicer.deleteClient,
                    request_deserializer=notification__service__pb2.ClientIdRequest.FromString,
                    response_serializer=notification__service__pb2.CommonResponse.SerializeToString,
            ),
            'isClientExists': grpc.unary_unary_rpc_method_handler(
                    servicer.isClientExists,
                    request_deserializer=notification__service__pb2.ClientIdRequest.FromString,
                    response_serializer=notification__service__pb2.isClientExistsResponse.SerializeToString,
            ),
            'timestampToEventOffset': grpc.unary_unary_rpc_method_handler(
                    servicer.timestampToEventOffset,
                    request_deserializer=notification__service__pb2.TimeToOffsetRequest.FromString,
                    response_serializer=notification__service__pb2.TimeToOffsetResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'notification_service.NotificationService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class NotificationService(object):
    """AirFlowService provides notification function rest endpoint of NotificationService for Notification Service component.
    Functions of NotificationService include:
    1.Send event.
    2.List events.
    """

    @staticmethod
    def sendEvent(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/notification_service.NotificationService/sendEvent',
            notification__service__pb2.SendEventRequest.SerializeToString,
            notification__service__pb2.SendEventsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def listEvents(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/notification_service.NotificationService/listEvents',
            notification__service__pb2.ListEventsRequest.SerializeToString,
            notification__service__pb2.ListEventsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def countEvents(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/notification_service.NotificationService/countEvents',
            notification__service__pb2.CountEventsRequest.SerializeToString,
            notification__service__pb2.CountEventsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def listAllEvents(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/notification_service.NotificationService/listAllEvents',
            notification__service__pb2.ListAllEventsRequest.SerializeToString,
            notification__service__pb2.ListEventsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def notify(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/notification_service.NotificationService/notify',
            notification__service__pb2.NotifyRequest.SerializeToString,
            notification__service__pb2.CommonResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def listMembers(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/notification_service.NotificationService/listMembers',
            notification__service__pb2.ListMembersRequest.SerializeToString,
            notification__service__pb2.ListMembersResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def notifyNewMember(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/notification_service.NotificationService/notifyNewMember',
            notification__service__pb2.NotifyNewMemberRequest.SerializeToString,
            notification__service__pb2.CommonResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def registerClient(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/notification_service.NotificationService/registerClient',
            notification__service__pb2.RegisterClientRequest.SerializeToString,
            notification__service__pb2.RegisterClientResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def deleteClient(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/notification_service.NotificationService/deleteClient',
            notification__service__pb2.ClientIdRequest.SerializeToString,
            notification__service__pb2.CommonResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def isClientExists(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/notification_service.NotificationService/isClientExists',
            notification__service__pb2.ClientIdRequest.SerializeToString,
            notification__service__pb2.isClientExistsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def timestampToEventOffset(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/notification_service.NotificationService/timestampToEventOffset',
            notification__service__pb2.TimeToOffsetRequest.SerializeToString,
            notification__service__pb2.TimeToOffsetResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



#
# Copyright 2022 The AI Flow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import inspect
import json
from typing import Any, Callable, Dict, List, Optional, Union

import yaml
from rich.box import ASCII_DOUBLE_HEAD
from rich.console import Console
from rich.syntax import Syntax
from rich.table import Table


class NotificationConsole(Console):
    """Notification rich console"""

    def print_as_json(self, data: Dict):
        """Renders dict as json text representation"""
        json_content = json.dumps(data)
        self.print(Syntax(json_content, "json", theme="ansi_dark"), soft_wrap=True)

    def print_as_yaml(self, data: Dict):
        """Renders dict as yaml text representation"""
        yaml_content = yaml.dump(data)
        self.print(Syntax(yaml_content, "yaml", theme="ansi_dark"), soft_wrap=True)

    def print_as_table(self, data: List[Dict]):
        """Renders list of dictionaries as table"""
        if not data:
            self.print("No data found")
            return

        table = SimpleTable(
            show_header=True,
        )
        for col in data[0].keys():
            table.add_column(col)

        for row in data:
            table.add_row(*[str(d) for d in row.values()])
        self.print(table)

    # pylint: disable=too-many-return-statements
    def _normalize_data(self, value: Any, output: str) -> Optional[Union[list, str, dict]]:
        if isinstance(value, (tuple, list)):
            if output == "table":
                return ",".join(self._normalize_data(x, output) for x in value)
            return [self._normalize_data(x, output) for x in value]
        if isinstance(value, dict) and output != "table":
            return {k: self._normalize_data(v, output) for k, v in value.items()}
        if inspect.isclass(value):
            return value.__name__
        if value is None:
            return None
        return str(value)

    def print_as(self, data: List[Union[Dict, Any]], output: str, mapper: Optional[Callable] = None):
        """Prints provided using format specified by output argument"""
        output_to_renderer = {
            "json": self.print_as_json,
            "yaml": self.print_as_yaml,
            "table": self.print_as_table,
        }
        renderer = output_to_renderer.get(output)
        if not renderer:
            raise ValueError(
                f"Unknown formatter: {output}. Allowed options: {list(output_to_renderer.keys())}"
            )

        if not all(isinstance(d, dict) for d in data) and not mapper:
            raise ValueError("To tabulate non-dictionary data you need to provide `mapper` function")

        if mapper:
            dict_data: List[Dict] = [mapper(d) for d in data]
        else:
            dict_data: List[Dict] = data
        dict_data = [{k: self._normalize_data(v, output) for k, v in d.items()} for d in dict_data]
        renderer(dict_data)


class SimpleTable(Table):
    """A rich Table with some default hardcoded for consistency."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.show_edge = kwargs.get("show_edge", False)
        self.pad_edge = kwargs.get("pad_edge", False)
        self.box = kwargs.get("box", ASCII_DOUBLE_HEAD)
        self.show_header = kwargs.get("show_header", False)
        self.title_style = kwargs.get("title_style", "bold green")
        self.title_justify = kwargs.get("title_justify", "left")
        self.caption = kwargs.get("caption", " ")

    def add_column(self, *args, **kwargs) -> None:  # pylint: disable=signature-differs
        """Add a column to the table. We use different default"""
        kwargs["overflow"] = kwargs.get("overflow", None)  # to avoid truncating
        super().add_column(*args, **kwargs)



#
# Copyright 2022 The AI Flow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import time
import uuid
from typing import Dict

import grpc
import logging
import threading
from abc import ABC, abstractmethod
from collections import deque

from notification_service.model.event import Event
from notification_service.model.member import Member
from notification_service.rpc.protobuf import notification_service_pb2_grpc
from notification_service.rpc.protobuf.notification_service_pb2 import NotifyRequest, Notify, NotifyNewMemberRequest
from notification_service.rpc.protobuf.notification_service_pb2_grpc import NotificationServiceStub
from notification_service.util.utils import sleep_and_detecting_running, member_to_proto

if not hasattr(time, 'time_ns'):
    time.time_ns = lambda: int(time.time() * 1e9)


class NotificationServerHaManager(ABC):

    @abstractmethod
    def start(self, server_uri, storage, ttl_ms, min_notify_interval, member_updated_condition):
        pass

    @abstractmethod
    def notify_others(self, event: Event):
        pass

    @abstractmethod
    def get_living_members(self):
        pass

    @abstractmethod
    def add_living_member(self, member: Member):
        pass

    @abstractmethod
    def stop(self):
        pass


class SimpleNotificationServerHaManager(NotificationServerHaManager):

    def __init__(self):
        self.server_uri = None
        self.storage = None
        self.ttl_ms = None
        self.cached_notify = deque()
        self.living_members = []
        self.member_connections = {}  # type: Dict[str, NotificationServiceStub]
        self.running = False
        self.heartbeat_thread = None
        self.notify_thread = None
        self.min_notify_interval_ms = None
        self.notified_others_after_start = False
        self.member_updated_condition = None
        self.uuid = str(uuid.uuid4())

    def start(self, server_uri, storage, ttl_ms, min_notify_interval_ms, member_updated_condition):
        self.server_uri = server_uri
        self.storage = storage
        self.ttl_ms = ttl_ms
        self.min_notify_interval_ms = min_notify_interval_ms
        self.member_updated_condition = member_updated_condition
        self.running = True
        self.heartbeat_thread = threading.Thread(target=self.start_heartbeat, daemon=True)
        self.heartbeat_thread.start()
        self.notify_thread = threading.Thread(target=self.start_notify, daemon=True)
        self.notify_thread.start()

    def start_heartbeat(self):
        while self.running:
            try:
                # do heartbeat
                self.storage.clear_dead_members(self.ttl_ms)
                self.storage.update_member(self.server_uri, self.uuid)
                self.living_members = self.storage.list_living_members(self.ttl_ms)
                if not self.notified_others_after_start:
                    for member in self.living_members:
                        if member.server_uri == self.server_uri:
                            continue
                        channel = grpc.insecure_channel(member.server_uri)
                        self.member_connections[member.server_uri] = \
                            notification_service_pb2_grpc.NotificationServiceStub(channel)
                        try:
                            self.member_connections[member.server_uri].notifyNewMember(
                                NotifyNewMemberRequest(member=member_to_proto(
                                    Member(1, self.server_uri, int(time.time_ns() / 1000000))
                                )))
                        except grpc.RpcError:
                            logging.error("Notify new member to '%s' failed." % member.server_uri,
                                          exc_info=True)
                    self.notified_others_after_start = True
            except Exception as e:
                logging.error("Exception thrown when send heartbeat to the HA storage.",
                              exc_info=True)
            sleep_and_detecting_running(self.ttl_ms / 2, lambda: self.running)

    def start_notify(self):
        while self.running:
            # update connections
            living_member_server_uris = set()
            for member in self.living_members:
                if member.server_uri == self.server_uri:
                    continue
                if member.server_uri not in self.member_connections:
                    try:
                        channel = grpc.insecure_channel(member.server_uri)
                        self.member_connections[member.server_uri] = \
                            notification_service_pb2_grpc.NotificationServiceStub(channel)
                    except Exception:
                        logging.error("Exception thrown when connect to another member: %s." %
                                      member.server_uri,
                                      exc_info=True)
                        continue
                living_member_server_uris.add(member.server_uri)
            for server_uri in list(self.member_connections.keys()):
                if server_uri not in living_member_server_uris:
                    del self.member_connections[server_uri]

            notifies = []
            while len(self.cached_notify) > 0:
                event = self.cached_notify.popleft()
                notify = Notify(key=event.key, namespace=event.namespace)
                notifies.append(notify)

            # notify others
            if len(notifies) > 0:
                for server_uri, stub in self.member_connections.items():
                    try:
                        stub.notify(NotifyRequest(notifies=notifies))
                    except Exception:
                        logging.error("Exception thrown when notify another member: %s." %
                                      server_uri,
                                      exc_info=True)
            sleep_and_detecting_running(self.min_notify_interval_ms, lambda: self.running, 100)

    def notify_others(self, event):
        self.cached_notify.append(event)

    def get_living_members(self):
        if len(self.living_members) == 0:
            self.living_members = self.storage.list_living_members(self.ttl_ms)
        return self.living_members

    def add_living_member(self, member: Member):
        if member.server_uri not in [member.server_uri for member in self.living_members]:
            self.living_members.append(member)

    def detach_member_from_cluster(self):
        self.storage.delete_member(self.server_uri)

    def stop(self):
        self.running = False
        self.detach_member_from_cluster()
        self.heartbeat_thread.join()
        self.notify_thread.join()



#!/usr/bin/env python
#
# Copyright 2022 The AI Flow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import argparse
import logging
import os

from notification_service.server import NotificationServerRunner
from notification_service.server_config import NotificationServerConfig
from notification_service.util import db
from notification_service.util.server_config import create_server_config


def _prepare_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('--generate-config-only', default=None, action='store_true',
                        help='Only generate notification server configuration file.')
    return parser.parse_args()


if __name__ == '__main__':
    logging.basicConfig(format='%(asctime)s - %(pathname)s[line:%(lineno)d] - %(levelname)s: %(message)s',
                        level=logging.INFO)
    if "NOTIFICATION_HOME" not in os.environ:
        os.environ["NOTIFICATION_HOME"] = os.environ["HOME"] + "/notification_service"
        logging.info("Set env variable NOTIFICATION_HOME to {}".format(os.environ["NOTIFICATION_HOME"]))
    if not os.path.exists(os.environ["NOTIFICATION_HOME"]):
        os.makedirs(os.environ["NOTIFICATION_HOME"])
    config_file = os.environ["NOTIFICATION_HOME"] + "/notification_server.yaml"

    args = _prepare_args()

    if not os.path.exists(config_file):
        create_server_config(os.environ.get("NOTIFICATION_HOME"), os.environ.copy())
        logging.info('Notification server config generated at {}.'.format(config_file))
    else:
        logging.info("Notification server config exists at {}".format(config_file))

    if args.generate_config_only:
        exit(0)

    config = NotificationServerConfig(config_file)
    db.create_all_tables(config.db_uri)
    ns = NotificationServerRunner(config_file=config_file)
    logging.info('notification server start(port:{}).'.format(ns.config.port))
    ns.start(is_block=True)



# Copyright 2022 The AI Flow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""Event commands"""
import functools
from argparse import Namespace
from datetime import datetime as dt
from queue import Queue
from typing import Callable, TypeVar, cast, List

from notification_service.client.embedded_notification_client import EmbeddedNotificationClient
from notification_service.client.notification_client import ListenerProcessor
from notification_service.model.event import Event

from ai_flow.common.util import time_utils
from notification_service.cli.simple_table import NotificationConsole


T = TypeVar("T", bound=Callable)  # pylint: disable=invalid-name


def check_arguments(f: T) -> T:

    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        if not args:
            raise ValueError("Args should be set")
        if not isinstance(args[0], Namespace):
            raise ValueError(
                "1st positional argument should be argparse.Namespace instance," f"but is {type(args[0])}."
            )

        if not args[0].server_uri:
            print("Argument --server-uri is not set. See `notification event {} -h`".format(args[0].subcommand))
            return
        return f(*args, **kwargs)

    return cast(T, wrapper)


@check_arguments
def list_events(args):
    """List events at the command line"""
    client = EmbeddedNotificationClient(server_uri=args.server_uri)
    offset = 0
    if args.begin_offset:
        offset = args.begin_offset
    elif args.begin_time:
        offset = client.time_to_offset(time_utils.timestamp_to_datetime(args.begin_time))
    events = client.list_events(key=args.key,
                                namespace=args.namespace,
                                sender=args.sender,
                                begin_offset=offset)
    NotificationConsole().print_as(
        data=events,
        output=args.output,
        mapper=lambda event: {
            "namespace": event.namespace,
            "key": event.key,
            "value": event.value,
            "sender": event.sender,
            "create_time": dt.fromtimestamp(event.create_time/1000).isoformat(),
            "context": event.context,
            "offset": event.offset,
        },
    )


@check_arguments
def count_events(args):
    """Count events at the command line"""
    client = EmbeddedNotificationClient(server_uri=args.server_uri)
    offset = 0
    if args.begin_offset:
        offset = args.begin_offset
    elif args.begin_time:
        offset = client.time_to_offset(time_utils.timestamp_to_datetime(args.begin_time))
    res = client.count_events(key=args.key,
                              namespace=args.namespace,
                              sender=args.sender,
                              begin_offset=offset)
    print(res[0])


@check_arguments
def listen_events(args):
    """Listen events at the command line"""
    class Processor(ListenerProcessor):
        def __init__(self, queue):
            self.queue = queue

        def process(self, events: List[Event]):
            for e in events:
                self.queue.put(e)

    client = EmbeddedNotificationClient(server_uri=args.server_uri,
                                        namespace=args.namespace)

    offset = 0
    if args.begin_offset:
        offset = args.begin_offset
    elif args.begin_time:
        offset = client.time_to_offset(time_utils.timestamp_to_datetime(args.begin_time))

    event_queue = Queue()
    registration_id = client.register_listener(listener_processor=Processor(event_queue),
                                               event_keys=[args.key, ],
                                               offset=offset)

    try:
        while True:
            print(event_queue.get())
    except KeyboardInterrupt:
        pass
    finally:
        client.unregister_listener(registration_id)


@check_arguments
def send_event(args):
    """Send an event at the command line"""
    client = EmbeddedNotificationClient(server_uri=args.server_uri,
                                        namespace=args.namespace,
                                        sender=args.sender)
    event = Event(
        key=args.key,
        value=args.value
    )
    event: Event = client.send_event(event)
    print("Successfully send event: {}.".format(event))
